/*
    cve-manager : CVE management tool
    Copyright (C) 2017-2024 Alexey Appolonov

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include    "data_preproc_kernel.h"

//-----------------------------------------------------------------------------
// Split versions of the kernel streams and place a flag indicating whether
// a particular stream is supported or not

bool PreprocessKernelStreams(tree::TSelectedData &selected_data,
    const bool is_supported_streams, string &msg){

    tree::TSelectedData selected_data_proc;

    try{
        for(const auto &entry : selected_data){
            for(const auto &item : entry){
                for(const auto &val : item){
                    if(val.empty()){
                        continue;
                        }
                    auto kernel_streams = Split(val, ",");
                    for(auto kernel_stream : kernel_streams){
                        Trim(kernel_stream, "{}");
                        tree::TSelectedDataEntryItem new_item_a, new_item_b;
                        new_item_a.push_back(kernel_stream);
                        new_item_b.push_back(is_supported_streams ? "1" : "0");
                        tree::TSelectedDataEntry new_entry;
                        new_entry.push_back(new_item_a);
                        new_entry.push_back(new_item_b);
                        selected_data_proc.push_back(new_entry);
                        }
                    }
                }
            }
        }
    catch(const exception &e){
        const string target = is_supported_streams ?
            TARGET_KERNEL_STREAMS_SUPPORTED :
            TARGET_KERNEL_STREAMS_EOL;
        msg = ErrMsg("Unable to preprocess " + target + " selected data", e);
        return false;
        }

    selected_data_proc.swap(selected_data);

    return true;
    }

//-----------------------------------------------------------------------------
// Split CVE string to CVE year and CVE number

bool PreprocessKernelFixes(tree::TSelectedData &selected_data, string &msg){

    tree::TSelectedData selected_data_proc;
    bool skip;

    // Form an error message
    auto wrong_fmt = [](const tree::TSelectedDataEntry &entry){
        return "Wrong format of the selected data : "
            "\"" + entry.Stringify() + "\"";
        };
    // - nested func

    try{
        for(auto &entry : selected_data){
            tree::TSelectedDataEntryItem cve_year_item, cve_num_item;
            tree::TSelectedDataEntry streams_items, fixes_items;
            skip = false;
            // Each entry consists of three items: 1) CVE ID, 2) kernel streams
            // and 3) fixed kernel versions
            if(entry.size() != 3){
                msg = wrong_fmt(entry);
                return false;
                }
            for(auto item = entry.begin(); item != entry.end(); item++){
                if(item->empty()){
                    msg = wrong_fmt(entry);
                    return false;
                    }
                // If this is a CVE ID
                if(item == entry.begin()){
                    if(item->size() != 1){
                        msg = wrong_fmt(entry);
                        return false;
                        }
                    const auto &el = item->front();
                    const auto cve_id_splitted = Split(el, "-");
                    // CVE ID consists of three parts
                    if(cve_id_splitted.size() != 3){
                        if(!msg.empty()){
                            msg += "; ";
                            }
                        msg += "Invalid CVE ID \"" + el + "\"";
                        skip = true;
                        break;
                        }
                    cve_year_item.push_back(cve_id_splitted.at(1));
                    cve_num_item.push_back(cve_id_splitted.at(2));
                    }
                // This is either a kernel stream or a fixed kernel version
                else{
                    tree::TSelectedDataEntry *items = NULL;
                    if(item == (entry.begin() + 1)){
                        items = &streams_items;
                        }
                    // There should be as many fixed versions as there are
                    // kernel streams
                    else{
                        if(item->size() != streams_items.size()){
                            msg = wrong_fmt(entry);
                            return false;
                            }
                        items = &fixes_items;
                        }
                    for(const auto &el : *item){
                        tree::TSelectedDataEntryItem new_item;
                        new_item.push_back(el);
                        items->push_back(new_item);
                        }
                    }
                }
            if(skip){
                continue;
                }
            for(size_t i = 0; i < streams_items.size(); i++){
                tree::TSelectedDataEntry new_entry;
                new_entry.push_back(cve_year_item);
                new_entry.push_back(cve_num_item);
                new_entry.push_back(streams_items[i]);
                new_entry.push_back(fixes_items[i]);
                selected_data_proc.push_back(new_entry);
                }
            }
        }
    catch(const exception &e){
        msg = ErrMsg(
            "Unable to preprocess " TARGET_KERNEL_FIXES " selected data", e);
        return false;
        }

    selected_data_proc.swap(selected_data);

    return true;
    }

//-----------------------------------------------------------------------------
